@@ -1,27 +1,19 @@
+import itertools
 from copy import copy
 from typing import Any, Optional, Union
 
 from pypika.enums import Dialects
 from pypika.queries import (
     CreateQueryBuilder,
+    Database,
+    DropQueryBuilder,
+    Selectable,
+    Table,
     Query,
     QueryBuilder,
-    Table,
 )
 from pypika.terms import ArithmeticExpression, Criterion, EmptyCriterion, Field, Function, Star, Term, ValueWrapper
-from pypika.utils import (
-    QueryException,
-    builder,
-)
-
-
-class SnowFlakeQueryBuilder(QueryBuilder):
-    QUOTE_CHAR = None
-    ALIAS_QUOTE_CHAR = '"'
-    QUERY_ALIAS_QUOTE_CHAR = ''
-
-    def __init__(self, **kwargs: Any) -> None:
-        super().__init__(dialect=Dialects.SNOWFLAKE, **kwargs)
+from pypika.utils import QueryException, builder, format_quotes
 
 
 class SnowflakeQuery(Query):
@@ -30,12 +22,61 @@ class SnowflakeQuery(Query):
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> SnowFlakeQueryBuilder:
-        return SnowFlakeQueryBuilder(**kwargs)
+    def _builder(cls, **kwargs: Any) -> "SnowflakeQueryBuilder":
+        return SnowflakeQueryBuilder(**kwargs)
+
+    @classmethod
+    def create_table(cls, table: Union[str, Table]) -> "SnowflakeCreateQueryBuilder":
+        return SnowflakeCreateQueryBuilder().create_table(table)
+
+    @classmethod
+    def drop_table(cls, table: Union[str, Table]) -> "SnowflakeDropQueryBuilder":
+        return SnowflakeDropQueryBuilder().drop_table(table)
+
+
+class SnowflakeQueryBuilder(QueryBuilder):
+    QUOTE_CHAR = None
+    ALIAS_QUOTE_CHAR = '"'
+    QUERY_ALIAS_QUOTE_CHAR = ''
+    QUERY_CLS = SnowflakeQuery
+
+    def __init__(self, **kwargs: Any) -> None:
+        super().__init__(dialect=Dialects.SNOWFLAKE, **kwargs)
+
+
+class SnowflakeCreateQueryBuilder(CreateQueryBuilder):
+    QUOTE_CHAR = None
+    QUERY_CLS = SnowflakeQuery
+
+    def __init__(self) -> None:
+        super().__init__(dialect=Dialects.SNOWFLAKE)
+
+
+class SnowflakeDropQueryBuilder(DropQueryBuilder):
+    QUOTE_CHAR = None
+    QUERY_CLS = SnowflakeQuery
+
+    def __init__(self) -> None:
+        super().__init__(dialect=Dialects.SNOWFLAKE)
+
+
+class MySQLQuery(Query):
+    """
+    Defines a query class for use with MySQL.
+    """
+
+    @classmethod
+    def _builder(cls, **kwargs: Any) -> "MySQLQueryBuilder":
+        return MySQLQueryBuilder(**kwargs)
+
+    @classmethod
+    def load(cls, fp: str) -> "MySQLLoadQueryBuilder":
+        return MySQLLoadQueryBuilder().load(fp)
 
 
 class MySQLQueryBuilder(QueryBuilder):
     QUOTE_CHAR = "`"
+    QUERY_CLS = MySQLQuery
 
     def __init__(self, **kwargs: Any) -> None:
         super().__init__(dialect=Dialects.MYSQL, wrap_set_operation_queries=False, **kwargs)
@@ -108,16 +149,18 @@ class MySQLQueryBuilder(QueryBuilder):
 
 
 class MySQLLoadQueryBuilder:
+    QUERY_CLS = MySQLQuery
+
     def __init__(self) -> None:
         self._load_file = None
         self._into_table = None
 
     @builder
-    def load(self, fp: str) -> "MySQLQueryBuilder":
+    def load(self, fp: str) -> "MySQLLoadQueryBuilder":
         self._load_file = fp
 
     @builder
-    def into(self, table: Union[str, Table]) -> "MySQLQueryBuilder":
+    def into(self, table: Union[str, Table]) -> "MySQLLoadQueryBuilder":
         self._into_table = table if isinstance(table, Table) else Table(table)
 
     def get_sql(self, *args: Any, **kwargs: Any) -> str:
@@ -142,21 +185,27 @@ class MySQLLoadQueryBuilder:
         return self.get_sql()
 
 
-class MySQLQuery(Query):
+class VerticaQuery(Query):
     """
-    Defines a query class for use with MySQL.
+    Defines a query class for use with Vertica.
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> "MySQLQueryBuilder":
-        return MySQLQueryBuilder(**kwargs)
+    def _builder(cls, **kwargs) -> "VerticaQueryBuilder":
+        return VerticaQueryBuilder(**kwargs)
 
     @classmethod
-    def load(cls, fp: str) -> "MySQLQueryBuilder":
-        return MySQLLoadQueryBuilder().load(fp)
+    def from_file(cls, fp: str) -> "VerticaCopyQueryBuilder":
+        return VerticaCopyQueryBuilder().from_file(fp)
+
+    @classmethod
+    def create_table(cls, table: Union[str, Table]) -> "VerticaCreateQueryBuilder":
+        return VerticaCreateQueryBuilder().create_table(table)
 
 
 class VerticaQueryBuilder(QueryBuilder):
+    QUERY_CLS = VerticaQuery
+
     def __init__(self, **kwargs: Any) -> None:
         super().__init__(dialect=Dialects.VERTICA, **kwargs)
         self._hint = None
@@ -175,6 +224,8 @@ class VerticaQueryBuilder(QueryBuilder):
 
 
 class VerticaCreateQueryBuilder(CreateQueryBuilder):
+    QUERY_CLS = VerticaQuery
+
     def __init__(self) -> None:
         super().__init__(dialect=Dialects.VERTICA)
         self._local = False
@@ -217,6 +268,8 @@ class VerticaCreateQueryBuilder(CreateQueryBuilder):
 
 
 class VerticaCopyQueryBuilder:
+    QUERY_CLS = VerticaQuery
+
     def __init__(self) -> None:
         self._copy_table = None
         self._from_file = None
@@ -251,48 +304,45 @@ class VerticaCopyQueryBuilder:
         return self.get_sql()
 
 
-class VerticaQuery(Query):
-    """
-    Defines a query class for use with Vertica.
-    """
-
-    @classmethod
-    def _builder(cls, **kwargs) -> VerticaQueryBuilder:
-        return VerticaQueryBuilder(**kwargs)
-
-    @classmethod
-    def from_file(cls, fp: str) -> VerticaCopyQueryBuilder:
-        return VerticaCopyQueryBuilder().from_file(fp)
-
-    @classmethod
-    def create_table(cls, table: Union[str, Table]) -> VerticaCreateQueryBuilder:
-        return VerticaCreateQueryBuilder().create_table(table)
-
-
-class OracleQueryBuilder(QueryBuilder):
-    QUOTE_CHAR = None
-
-    def __init__(self, **kwargs: Any) -> None:
-        super().__init__(dialect=Dialects.ORACLE, **kwargs)
-
-    def get_sql(self, *args: Any, **kwargs: Any) -> str:
-        return super().get_sql(*args, groupby_alias=False, **kwargs)
-
-
 class OracleQuery(Query):
     """
     Defines a query class for use with Oracle.
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> OracleQueryBuilder:
+    def _builder(cls, **kwargs: Any) -> "OracleQueryBuilder":
         return OracleQueryBuilder(**kwargs)
 
 
-class PostgreQueryBuilder(QueryBuilder):
-    ALIAS_QUOTE_CHAR = '"'
+class OracleQueryBuilder(QueryBuilder):
+    QUOTE_CHAR = None
+    QUERY_CLS = OracleQuery
 
-    def __init__(self, **kwargs: str) -> None:
+    def __init__(self, **kwargs: Any) -> None:
+        super().__init__(dialect=Dialects.ORACLE, **kwargs)
+
+    def get_sql(self, *args: Any, **kwargs: Any) -> str:
+        # Oracle does not support group by a field alias
+        # Note: set directly in kwargs as they are re-used down the tree in the case of subqueries!
+        kwargs['groupby_alias'] = False
+        return super().get_sql(*args, **kwargs)
+
+
+class PostgreSQLQuery(Query):
+    """
+    Defines a query class for use with PostgreSQL.
+    """
+
+    @classmethod
+    def _builder(cls, **kwargs) -> "PostgreSQLQueryBuilder":
+        return PostgreSQLQueryBuilder(**kwargs)
+
+
+class PostgreSQLQueryBuilder(QueryBuilder):
+    ALIAS_QUOTE_CHAR = '"'
+    QUERY_CLS = PostgreSQLQuery
+
+    def __init__(self, **kwargs: Any) -> None:
         super().__init__(dialect=Dialects.POSTGRESQL, **kwargs)
         self._returns = []
         self._return_star = False
@@ -306,14 +356,14 @@ class PostgreQueryBuilder(QueryBuilder):
 
         self._distinct_on = []
 
-    def __copy__(self) -> "PostgreQueryBuilder":
+    def __copy__(self) -> "PostgreSQLQueryBuilder":
         newone = super().__copy__()
         newone._returns = copy(self._returns)
         newone._on_conflict_do_updates = copy(self._on_conflict_do_updates)
         return newone
 
     @builder
-    def distinct_on(self, *fields: Union[str, Term]) -> "PostgreQueryBuilder":
+    def distinct_on(self, *fields: Union[str, Term]) -> "PostgreSQLQueryBuilder":
         for field in fields:
             if isinstance(field, str):
                 self._distinct_on.append(Field(field))
@@ -321,7 +371,7 @@ class PostgreQueryBuilder(QueryBuilder):
                 self._distinct_on.append(field)
 
     @builder
-    def on_conflict(self, *target_fields: Union[str, Term]) -> "PostgreQueryBuilder":
+    def on_conflict(self, *target_fields: Union[str, Term]) -> "PostgreSQLQueryBuilder":
         if not self._insert_table:
             raise QueryException("On conflict only applies to insert query")
 
@@ -334,13 +384,15 @@ class PostgreQueryBuilder(QueryBuilder):
                 self._on_conflict_fields.append(target_field)
 
     @builder
-    def do_nothing(self) -> "PostgreQueryBuilder":
+    def do_nothing(self) -> "PostgreSQLQueryBuilder":
         if len(self._on_conflict_do_updates) > 0:
             raise QueryException("Can not have two conflict handlers")
         self._on_conflict_do_nothing = True
 
     @builder
-    def do_update(self, update_field: Union[str, Field], update_value: Optional[Any] = None) -> "PostgreQueryBuilder":
+    def do_update(
+        self, update_field: Union[str, Field], update_value: Optional[Any] = None
+    ) -> "PostgreSQLQueryBuilder":
         if self._on_conflict_do_nothing:
             raise QueryException("Can not have two conflict handlers")
 
@@ -357,7 +409,7 @@ class PostgreQueryBuilder(QueryBuilder):
             self._on_conflict_do_updates.append((field, None))
 
     @builder
-    def where(self, criterion: Criterion) -> "PostgreQueryBuilder":
+    def where(self, criterion: Criterion) -> "PostgreSQLQueryBuilder":
         if not self._on_conflict:
             return super().where(criterion)
 
@@ -380,6 +432,10 @@ class PostgreQueryBuilder(QueryBuilder):
         else:
             raise QueryException('Can not have fieldless ON CONFLICT WHERE')
 
+    @builder
+    def using(self, table: Union[Selectable, str]) -> "QueryBuilder":
+        self._using.append(table)
+
     def _distinct_sql(self, **kwargs: Any) -> str:
         if self._distinct_on:
             return "DISTINCT ON({distinct_on}) ".format(
@@ -441,7 +497,7 @@ class PostgreQueryBuilder(QueryBuilder):
         return ''
 
     @builder
-    def returning(self, *terms: Any) -> "PostgreQueryBuilder":
+    def returning(self, *terms: Any) -> "PostgreSQLQueryBuilder":
         for term in terms:
             if isinstance(term, Field):
                 self._return_field(term)
@@ -458,7 +514,12 @@ class PostgreQueryBuilder(QueryBuilder):
         for field in term.fields_():
             if not any([self._insert_table, self._update_table, self._delete_from]):
                 raise QueryException("Returning can't be used in this query")
-            if field.table not in {self._insert_table, self._update_table} and term not in self._from:
+
+            table_is_insert_or_update_table = field.table in {self._insert_table, self._update_table}
+            join_tables = set(itertools.chain.from_iterable([j.criterion.tables_ for j in self._joins]))
+            join_and_base_tables = set(self._from) | join_tables
+            table_not_base_or_join = bool(term.tables_ - join_and_base_tables)
+            if not table_is_insert_or_update_table and table_not_base_or_join:
                 raise QueryException("You can't return from other tables")
 
     def _set_returns_for_star(self) -> None:
@@ -504,7 +565,7 @@ class PostgreQueryBuilder(QueryBuilder):
     def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
         self._set_kwargs_defaults(kwargs)
 
-        querystring = super(PostgreQueryBuilder, self).get_sql(with_alias, subquery, **kwargs)
+        querystring = super(PostgreSQLQueryBuilder, self).get_sql(with_alias, subquery, **kwargs)
 
         querystring += self._on_conflict_sql(**kwargs)
         querystring += self._on_conflict_action_sql(**kwargs)
@@ -515,38 +576,43 @@ class PostgreQueryBuilder(QueryBuilder):
         return querystring
 
 
-class PostgreSQLQuery(Query):
-    """
-    Defines a query class for use with PostgreSQL.
-    """
-
-    @classmethod
-    def _builder(cls, **kwargs) -> PostgreQueryBuilder:
-        return PostgreQueryBuilder(**kwargs)
-
-
 class RedshiftQuery(Query):
     """
     Defines a query class for use with Amazon Redshift.
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> "QueryBuilder":
-        return QueryBuilder(dialect=Dialects.REDSHIFT, **kwargs)
+    def _builder(cls, **kwargs: Any) -> "RedShiftQueryBuilder":
+        return RedShiftQueryBuilder(dialect=Dialects.REDSHIFT, **kwargs)
+
+
+class RedShiftQueryBuilder(QueryBuilder):
+    QUERY_CLS = RedshiftQuery
+
+
+class MSSQLQuery(Query):
+    """
+    Defines a query class for use with Microsoft SQL Server.
+    """
+
+    @classmethod
+    def _builder(cls, **kwargs: Any) -> "MSSQLQueryBuilder":
+        return MSSQLQueryBuilder(**kwargs)
 
 
 class MSSQLQueryBuilder(QueryBuilder):
+    QUERY_CLS = MSSQLQuery
+
     def __init__(self, **kwargs: Any) -> None:
         super().__init__(dialect=Dialects.MSSQL, **kwargs)
-        self._top = None
+        self._top: Union[int, None] = None
+        self._top_with_ties: bool = False
+        self._top_percent: bool = False
 
     @builder
-    def top(self, value: Union[str, int]) -> "MSSQLQueryBuilder":
+    def top(self, value: Union[str, int], percent: bool = False, with_ties: bool = False) -> "MSSQLQueryBuilder":
         """
         Implements support for simple TOP clauses.
-
-        Does not include support for PERCENT or WITH TIES.
-
         https://docs.microsoft.com/en-us/sql/t-sql/queries/top-transact-sql?view=sql-server-2017
         """
         try:
@@ -554,6 +620,11 @@ class MSSQLQueryBuilder(QueryBuilder):
         except ValueError:
             raise QueryException("TOP value must be an integer")
 
+        if percent and not (0 <= int(value) <= 100):
+            raise QueryException("TOP value must be between 0 and 100 when `percent`" " is specified")
+        self._top_percent: bool = percent
+        self._top_with_ties: bool = with_ties
+
     @builder
     def fetch_next(self, limit: int) -> "MSSQLQueryBuilder":
         # Overridden to provide a more domain-specific API for T-SQL users
@@ -577,13 +648,21 @@ class MSSQLQueryBuilder(QueryBuilder):
         return querystring
 
     def get_sql(self, *args: Any, **kwargs: Any) -> str:
-        return super().get_sql(*args, groupby_alias=False, **kwargs)
+        # MSSQL does not support group by a field alias.
+        # Note: set directly in kwargs as they are re-used down the tree in the case of subqueries!
+        kwargs['groupby_alias'] = False
+        return super().get_sql(*args, **kwargs)
 
     def _top_sql(self) -> str:
+        _top_statement: str = ""
         if self._top:
-            return "TOP ({}) ".format(self._top)
-        else:
-            return ""
+            _top_statement = f"TOP ({self._top}) "
+            if self._top_percent:
+                _top_statement = f"{_top_statement}PERCENT "
+            if self._top_with_ties:
+                _top_statement = f"{_top_statement}WITH TIES "
+
+        return _top_statement
 
     def _select_sql(self, **kwargs: Any) -> str:
         return "SELECT {distinct}{top}{select}".format(
@@ -593,17 +672,45 @@ class MSSQLQueryBuilder(QueryBuilder):
         )
 
 
-class MSSQLQuery(Query):
+class ClickHouseQuery(Query):
     """
-    Defines a query class for use with Microsoft SQL Server.
+    Defines a query class for use with Yandex ClickHouse.
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> MSSQLQueryBuilder:
-        return MSSQLQueryBuilder(**kwargs)
+    def _builder(cls, **kwargs: Any) -> "ClickHouseQueryBuilder":
+        return ClickHouseQueryBuilder(
+            dialect=Dialects.CLICKHOUSE, wrap_set_operation_queries=False, as_keyword=True, **kwargs
+        )
+
+    @classmethod
+    def drop_database(self, database: Union[Database, str]) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_database(database)
+
+    @classmethod
+    def drop_table(self, table: Union[Table, str]) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_table(table)
+
+    @classmethod
+    def drop_dictionary(self, dictionary: str) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_dictionary(dictionary)
+
+    @classmethod
+    def drop_quota(self, quota: str) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_quota(quota)
+
+    @classmethod
+    def drop_user(self, user: str) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_user(user)
+
+    @classmethod
+    def drop_view(self, view: str) -> "ClickHouseDropQueryBuilder":
+        return ClickHouseDropQueryBuilder().drop_view(view)
 
 
 class ClickHouseQueryBuilder(QueryBuilder):
+    QUERY_CLS = ClickHouseQuery
+
     @staticmethod
     def _delete_sql(**kwargs: Any) -> str:
         return 'ALTER TABLE'
@@ -628,23 +735,41 @@ class ClickHouseQueryBuilder(QueryBuilder):
         )
 
 
-class ClickHouseQuery(Query):
-    """
-    Defines a query class for use with Yandex ClickHouse.
-    """
+class ClickHouseDropQueryBuilder(DropQueryBuilder):
+    QUERY_CLS = ClickHouseQuery
 
-    @classmethod
-    def _builder(cls, **kwargs: Any) -> QueryBuilder:
-        return ClickHouseQueryBuilder(
-            dialect=Dialects.CLICKHOUSE, wrap_set_operation_queries=False, as_keyword=True, **kwargs
-        )
+    def __init__(self):
+        super().__init__(dialect=Dialects.CLICKHOUSE)
+        self._cluster_name = None
+
+    @builder
+    def drop_dictionary(self, dictionary: str) -> "ClickHouseDropQueryBuilder":
+        super()._set_target('DICTIONARY', dictionary)
+
+    @builder
+    def drop_quota(self, quota: str) -> "ClickHouseDropQueryBuilder":
+        super()._set_target('QUOTA', quota)
+
+    @builder
+    def on_cluster(self, cluster: str) -> "ClickHouseDropQueryBuilder":
+        if self._cluster_name:
+            raise AttributeError("'DropQuery' object already has attribute cluster_name")
+        self._cluster_name = cluster
+
+    def get_sql(self, **kwargs: Any) -> str:
+        query = super().get_sql(**kwargs)
+
+        if self._drop_target_kind != "DICTIONARY" and self._cluster_name is not None:
+            query += " ON CLUSTER " + format_quotes(self._cluster_name, super().QUOTE_CHAR)
+
+        return query
 
 
 class SQLLiteValueWrapper(ValueWrapper):
-    def get_value_sql(self, *args: Any, **kwargs: Any) -> str:
+    def get_value_sql(self, **kwargs: Any) -> str:
         if isinstance(self.value, bool):
             return "1" if self.value else "0"
-        return super().get_value_sql(*args, **kwargs)
+        return super().get_value_sql(**kwargs)
 
 
 class SQLLiteQuery(Query):
@@ -653,5 +778,23 @@ class SQLLiteQuery(Query):
     """
 
     @classmethod
-    def _builder(cls, **kwargs: Any) -> QueryBuilder:
-        return QueryBuilder(dialect=Dialects.SQLLITE, wrapper_cls=SQLLiteValueWrapper, **kwargs)
+    def _builder(cls, **kwargs: Any) -> "SQLLiteQueryBuilder":
+        return SQLLiteQueryBuilder(**kwargs)
+
+
+class SQLLiteQueryBuilder(QueryBuilder):
+    QUERY_CLS = SQLLiteQuery
+
+    def __init__(self, **kwargs: Any) -> None:
+        super().__init__(dialect=Dialects.SQLLITE, wrapper_cls=SQLLiteValueWrapper, **kwargs)
+        self._insert_or_replace = False
+
+    @builder
+    def insert_or_replace(self, *terms: Any) -> "SQLLiteQueryBuilder":
+        self._apply_terms(*terms)
+        self._replace = True
+        self._insert_or_replace = True
+
+    def _replace_sql(self, **kwargs: Any) -> str:
+        prefix = "INSERT OR " if self._insert_or_replace else ""
+        return prefix + super()._replace_sql(**kwargs)
